from santic_validation import MethodType, SanticModel


class UserSchema(SanticModel):
    name: str
    age: MethodType[int]


class TestFields:
    async def test_positive_method_type(self):
        data = {"name": "max", "age": 10}
        user = UserSchema(**data)
        assert user.dict() == data
        assert user._context == {}

    async def test_negative_method_type(self):
        try:
            UserSchema(name="Max", age="10")
        except Exception as err:
            assert err.errors() == [
                {"loc": ("age",), "msg": "Expected type int", "type": "type_error"}
            ]

        class UserSchema2(SanticModel):
            name: str
            age: MethodType

        try:
            UserSchema2(name="Max", age="10")
        except Exception as err:
            assert err.errors() == [
                {"loc": ("age",), "msg": "Value type not found", "type": "value_error"}
            ]



from inspect import iscoroutinefunction
from typing import Any, Callable, Dict, TypedDict, TypeVar, Union

from .fields import SanticModel

Schema = TypeVar("Schema")


class MethodData(TypedDict):
    method: Callable
    params: Dict[str, str]


async def async_validate_by_method(
    schema: Schema,
    data: Dict[str, Dict[str, Union[Callable, Dict[str, str]]]],
    exceptions: Union[Any],
    skip_on_fail: bool = False,
):
    error = {}
    for field_name, method_data in data.items():
        try:
            value = await method_data["method"](**method_data["params"])
            setattr(schema, field_name, value)
        except exceptions:
            error[field_name] = "Invalid value"
            if skip_on_fail:
                return error

    return True if not error else False, error or None


async def validate_method_fields(schema: SanticModel, replace_value=False):
    method_fields = getattr(schema, "_method_fields", None)
    if not method_fields:
        return

    for name in method_fields:
        method, params = (
            schema.__fields__[name].field_info.extra.get("method"),
            getattr(schema, f"{name}_method_params", {}),
        )
        if iscoroutinefunction(method):
            value = await method(**params)
        else:
            value = method(**params)

        if replace_value:
            setattr(schema, name, value)



from pydantic import BaseModel, Field, ValidationError

# isort: off
from santic_validation import (
    MethodType,
    SanticModel,
    async_validate_by_method,
    validate_method_fields,
)

# isort: on


class PlanetSchema(BaseModel):
    name: str
    age: int


class TestAsyncValidate:
    @staticmethod
    async def age_to_string(value) -> str:
        return str(value)

    async def test_positive_async_validate_by_method(self):
        schema = PlanetSchema(name="Anna", age=20)
        assert isinstance(schema.age, int)

        is_valid, errors = await async_validate_by_method(
            schema,
            {
                "age": {
                    "method": self.age_to_string,
                    "params": {"value": schema.age},
                }
            },
            exceptions=(ValidationError,),
        )

        assert is_valid
        assert errors is None
        assert isinstance(schema.age, str)

    async def test_negative_async_validate_by_method(self):
        schema = PlanetSchema(name="Anna", age=20)
        assert isinstance(schema.age, int)

        is_valid, errors = await async_validate_by_method(
            schema,
            {
                "age_test": {
                    "method": self.age_to_string,
                    "params": {"value": schema.age},
                }
            },
            exceptions=(
                ValidationError,
                ValueError,
            ),
        )

        assert is_valid is False
        assert errors == {"age_test": "Invalid value"}

    async def test_validate_method_fields(self):
        class Age(object):
            value: int

            def __init__(self, value):
                self.value = value

            @classmethod
            def to_instance(cls, age: int):
                return cls(value=age)

        class UserSchema(SanticModel):
            name: str
            age: MethodType[int] = Field(method=Age.to_instance)

            @property
            def age_method_params(self):
                return {"age": self.age}

        data = {"name": "Zack", "age": 10}
        user = UserSchema(**data)

        await validate_method_fields(user)
        assert data == user.dict()

        await validate_method_fields(user, replace_value=True)
        assert user.name == data["name"]
        assert isinstance(user.age, Age) is True

        async def multiply(value):
            return value * 2

        class PersonSchema(SanticModel):
            name: str
            age: MethodType[int] = Field(method=multiply)

            @property
            def age_method_params(self):
                return {"value": self.age}

        data = {"name": "Zack", "age": 10}
        user = PersonSchema(**data)

        await validate_method_fields(user)
        assert data == user.dict()

        data["age"] = data["age"] * 2
        await validate_method_fields(user, replace_value=True)
        assert data == user.dict()



from functools import wraps
from inspect import isawaitable
from typing import Dict, Optional, Type, get_args, get_origin, get_type_hints

from pydantic import ValidationError
from sanic import Request
from sanic.exceptions import SanicException

from .fields import SanticModel
from .utils import validate_method_fields

ARRAY_TYPES = {list, tuple}


def has_array_type(hint_value) -> Optional[bool]:
    if get_origin(hint_value) in ARRAY_TYPES:
        return True

    list_args = get_args(hint_value)
    for args in list_args:
        if get_origin(args) in ARRAY_TYPES:
            return True


def clean_data(schema, raw_data) -> Dict[str, any]:
    hints = get_type_hints(schema)
    data = {}

    for hint_key, hint_value in hints.items():
        is_array = has_array_type(hint_value)
        if is_array:
            value = raw_data.getlist(hint_key)
        else:
            value = raw_data.get(hint_key)

        if value and (hint_value == int or int in get_args(hint_value)):
            if isinstance(value, (list, tuple)):
                value = map(lambda x: int(x) if x.isdigit() else x, value)
            else:
                value = int(value) if value.isdigit() else value

        data[hint_key] = value

    return data


def validate_schema(
    body: Optional[Type[object]] = None,
    query: Optional[Type[object]] = None,
    method_replace_value=False,
):
    """
    Simple validation
    """

    def decorator(f):
        @wraps(f)
        async def decorated_function(*args, **kwargs):
            if args and isinstance(args[0], Request):
                request: Request = args[0]
            elif len(args) > 1:
                request: Request = args[1]
            else:
                raise SanicException("Request could not be found")

            try:
                if query:
                    cleaned_data = clean_data(
                        query,
                        request.args,
                    )
                    kwargs["query"] = query(
                        **cleaned_data,
                    )
                if body:
                    if request.headers["content-type"] == "application/json":
                        cleaned_data = request.json
                    else:
                        cleaned_data = clean_data(
                            body,
                            request.form,
                        )

                    kwargs["body"] = body(**cleaned_data)
                    if isinstance(kwargs["body"], SanticModel):
                        if kwargs.get("query"):
                            kwargs["body"]._context = {"query": kwargs["query"]}

                        await validate_method_fields(
                            kwargs["body"], replace_value=method_replace_value
                        )

            except ValidationError as err:
                raise SanicException(
                    "Validation error",
                    context={
                        "detail": {
                            error["loc"][0]: error["msg"] for error in err.errors()
                        }
                    },
                    status_code=400,
                )

            retval = f(*args, **kwargs)
            if isawaitable(retval):
                retval = await retval
            return retval

        return decorated_function

    return decorator



from typing import Dict, Generic, TypeVar, Union, no_type_check

from pydantic import BaseModel, PrivateAttr
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass

ValueType = TypeVar("value")


class MethodType(Generic[ValueType]):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v, field: ModelField, values):
        value_type = field.sub_fields[0] if field.sub_fields else None
        if not value_type:
            raise ValueError("Value type not found")

        if not isinstance(v, value_type.type_):
            raise TypeError(f"Expected type {value_type.type_.__name__}")

        return v


class Meta(ModelMetaclass):
    @no_type_check  # noqa C901
    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        setattr(cls, "_method_fields", [])
        for name, model_field in cls.__fields__.items():
            if model_field.type_ == MethodType:
                cls._method_fields.append(name)

        return cls


class SanticModel(BaseModel, metaclass=Meta):
    _context: Dict[Union[str, int, float], any] = PrivateAttr(default={})



from typing import List, Optional

import pytest
from pydantic import BaseModel, Field
from sanic import Sanic
from sanic.response import json, json_dumps

from santic_validation import MethodType, SanticModel, validate_schema


class BodySchema(BaseModel):
    name: str
    age: int
    list_address: List[str]
    list_ids: Optional[List[int]]


class QuerySchema(BaseModel):
    page: int


def address_number(value: str):
    return f"Number: {value}"


def make_address(value: str):
    return f"Address: {value}"


class AddressSchema(SanticModel):
    number: MethodType[int] = Field(method=address_number)
    address: MethodType[str] = Field(method=make_address)

    @property
    def address_method_params(self):
        query: QuerySchema = self._context.get("query")
        return {"value": f"{self.address} ({query.location})"}

    @property
    def number_method_params(self):
        query: QuerySchema = self._context.get("query")
        return {"value": f"{self.number} ({query.location})"}


class QueryAddressSchema(BaseModel):
    location: str


@pytest.fixture
def app():
    sanic_app = Sanic(name="TestingApp")

    @sanic_app.route("/post", methods=["POST"])
    @validate_schema(body=BodySchema, query=QuerySchema)
    async def post_handler(request, body, query):
        return json(
            {
                "body": body.dict(),
                "query": query.dict(),
            }
        )

    @sanic_app.route("/post_address", methods=["POST"])
    @validate_schema(body=AddressSchema, query=QueryAddressSchema)
    async def post_handler(request, body, query):
        return json(
            {
                "body": body.dict(),
                "query": query.dict(),
            }
        )

    @sanic_app.route("/post_address_replace", methods=["POST"])
    @validate_schema(
        body=AddressSchema,
        query=QueryAddressSchema,
        method_replace_value=True,
    )
    async def post_handler(request, body, query):
        return json(
            {
                "body": body.dict(),
                "query": query.dict(),
            }
        )

    return sanic_app


class TestValidateSchema:
    def test_positive_validate_schema(self, app):
        # Test request.json
        payload = {
            "name": "Anna",
            "age": 20,
            "list_address": ["address 1", "address 2"],
            "list_ids": [1, 2, 3],
        }
        query = {"page": 10}
        headers = {"content-type": "application/json"}

        response = app.test_client.post(
            "/post",
            data=json_dumps(payload),
            params=query,
            headers=headers,
        )[1]

        assert response.status_code == 200
        assert response.json == {
            "body": payload,
            "query": query,
        }

        # Test request.form
        payload = {
            "name": "Anna",
            "age": 20,
            "list_address": ["address 1", "address 2"],
            "list_ids": [1, 2, 3],
        }
        headers = {"content-type": "application/x-www-form-urlencoded"}
        data = payload
        request, response = app.test_client.post(
            "/post",
            data=data,
            params=query,
            headers=headers,
        )

        assert response.status_code == 200
        assert response.json == {
            "body": data,
            "query": query,
        }

    def test_negative_validate_schema(self, app):
        # test request.json
        payload = {"any": 1}
        headers = {"content-type": "application/json"}
        response = app.test_client.post(
            "/post",
            data=payload,
            headers=headers,
        )[1]

        assert response.status_code == 400

        # test request.form
        headers = {"content-type": "application/x-www-form-urlencoded"}
        response = app.test_client.post(
            "/post",
            data=payload,
            params={"page": 10},
            headers=headers,
        )[1]

        assert response.status_code == 400

    def test_validate_method_fields(self, app):
        # test validate methods
        data = {"address": "my-address", "number": 101}
        query = {"location": "Earth"}

        headers = {"content-type": "application/json"}
        request, response = app.test_client.post(
            "/post_address",
            data=json_dumps(data),
            params=query,
            headers=headers,
        )

        assert response.status_code == 200
        assert response.json == {
            "body": data,
            "query": query,
        }

        headers = {"content-type": "application/x-www-form-urlencoded"}
        request, response = app.test_client.post(
            "/post_address",
            data=data,
            params=query,
            headers=headers,
        )
        assert response.status_code == 200
        assert response.json == {
            "body": data,
            "query": query,
        }

        # test with replace value
        data = {"address": "my-address", "number": 101}
        query = {"location": "Earth"}

        headers = {"content-type": "application/json"}
        request, response = app.test_client.post(
            "/post_address_replace",
            data=json_dumps(data),
            params=query,
            headers=headers,
        )

        resp_data = data.copy()
        resp_data["address"] = f"Address: {data['address']} ({query['location']})"
        resp_data["number"] = f"Number: {data['number']} ({query['location']})"

        assert response.status_code == 200
        assert response.json == {
            "body": resp_data,
            "query": query,
        }

        data = {"address": "my-address", "number": 101}
        query = {"location": "Earth"}

        headers = {"content-type": "application/x-www-form-urlencoded"}
        request, response = app.test_client.post(
            "/post_address_replace",
            data=data,
            params=query,
            headers=headers,
        )

        resp_data = data.copy()
        resp_data["address"] = f"Address: {data['address']} ({query['location']})"
        resp_data["number"] = f"Number: {data['number']} ({query['location']})"

        assert response.status_code == 200
        assert response.json == {
            "body": resp_data,
            "query": query,
        }
